//
//                        ----------------------
// Sharpliner Calibration ---> SDCC version <---
//                        ----------------------
//
// Sharp distance sensor linear- and filterizer (tm) using a PIC12F683.
// (C) 11/2009 ASkr, www.askrprojects.net
//
// Note: data.c is 'compiled' via #include "data.c"!
//       Otherwise SDCC is not able to fit this into 12F683's memory!
//
// All functions (SetPWM(), GetAD()) have been put in main() directly.
// This might look ugly but save some memory...
//
// +----------------------------+
// | Code uses TABs. Use TAB=2. |
// +----------------------------+
//

#define VERSIONH 00
#define VERSIONL 07

// changes/revision history:
//
// v0.7
//  - moved the MSBs from code to EEPROM memory
//  - all EEPROM values (samples, avgs, address, ...) moved to (hard) code
//  
//
// v0.6
//  - implemented SPI like interface (PISS: proprietary interconnection serial system)
//  - ...
//
// v0.5
//  - implemented SPI like interface (PISS: proprietary interconnection serial system)
//  - removed functions getAD() and setPWM(); now in main(), saves flash memory
//  - ...


#include <pic/pic12f683.h>
#include <math.h>





//--------------------------
// CHANGE THIS TO YOUR NEEDS
#define cADDRESS	0									// bus address (0-19)
#define cAVG			5									// # of moving averages to perform (0-..; default: 5) Large values may not fit in RAM!
#define cSAMPLES 25									// # of samples per average (0-63; default: 25)
#define cDOCALIB	0									// 0->standard operation; 1->calibration (OUT:IN = 1:1)






//------------------------
// defines (do not change)
#define cPULSES 255-(cADDRESS*11);	// number of T1CKI pulses until an interrput occurs


//------------------------
// chip configs
// This is from the HiTech PICC Pro version
//__CONFIG(0x33f4);
//__EEPROM_DATA(cAVG,0xff,0xff,0xff,0xff,0xff,0xff,0xff);
__code unsigned int  __at 0x2007 __CONFIG=0x33D4;

// this moved to "data.c", which is generated by the octave script
//__code unsigned char __at(0x2100) __EEPROM[] = {9,8,7,6,5,4,3,2,1,0};



//------------------------
// prototypes
void Init(void);
void SetPWM(unsigned int val);



//------------------------
// little helpers
#define BANK0 bcf _STATUS,5
#define BANK1 bsf _STATUS,5


//------------------------
// globals
#include "data.c"													// calibration data; #include saves some memory, here

unsigned int gAvg[cAVG];    	 						// array of moving averages
volatile unsigned int gMavg;							// filtered A/D value; needs to be global for interrupt access


//****************************************************************************
//*** Init
//***                     
//***
//****************************************************************************
void Init()
{
	_asm
			BANK1
				movlw		b'01110000'		// switch to full speed
				iorwf		_OSCCON,f
				
				movlw		b'00000000'		// pull-ups in WPU; Presc. to Timer0 1:1 (unused)
				movwf		_OPTION_REG
				
				movlw		b'00100001'		// pulls-ups: GP0 (DATA), GP5 (CLK)
				movwf		_WPU
				
			BANK0
				clrf		_GPIO

			BANK1
				movlw		b'01011010'		// AN3 (sens in) and AN1 (pulse trig) analog, AD: 1/16 fosc
				movwf		_ANSEL

				movlw		b'00110011'		// inputs: AN3/GP4 (sensor), GP0 (DATA), GP1 (PULSETRIG), GP5 (CLK)
				movwf		_TRISIO
	
			BANK0
				movlw		b'10001101'		// ADC enabled, right, AN3, ref is VDD
				movwf		_ADCON0
	
				movlw		b'00001100'		// PWM mode, active high; b5,4 -> PWM LSBs
				movwf		_CCP1CON

				movlw		0							// PWM value (8 MSBs)
				movwf		_CCPR1L

			BANK1	
				movlw		255						// (see T2) (255+1)*500ns => 128us => ~7.8kHz
				movwf		_PR2
				
				movlw		b'10100111'		// VREN on; Uref=VR[0:3]/24 * Vdd; 7digits -> ~1V
				movwf		_VRCON
				
			BANK0
				movlw		b'00000100'		// Timer2 on, post 1:1, pre 1:1 => 500ns
				movwf		_T2CON
	
				movlw		b'00000110'		// Timer1 as counter T1CKI pin; off, 1:1
				movwf		_T1CON
				
				movlw		b'11000000'		// GIE on, PIE on
				movwf		_INTCON
				
				movlw		b'00000100'		// comparator on with + on CVref, non-inverted
				movwf		_CMCON0

				// !BANK!
				// movlw		b'00000000'		// 
				// movwf		_CMCON1
				
			BANK1
				movlw		b'00000000'		// we'll start with all INTs off...
				movwf		_PIE1

			BANK0										// just in case...
	_endasm;
		
	return;
}


//****************************************************************************
//*** SetPWM
//***
//***
//****************************************************************************
void SetPWM(unsigned int val)
{
	CCPR1L=(val & 0x3fc)>>2;
	CCP1CON=(CCP1CON & 0b00001100) | ((val & 0x03)<<4);
}


//****************************************************************************
//*** EnableCounterInt
//***
//*** Enables the pulse counting interrupt on T1CKI.
//*** Sets TMR1L/H to the corresponding pulse count number, depending
//*** on the sensor's bus address.
//*** Disables pulse trigger interrupt on CIN-.
//****************************************************************************
void EnableCounterInt()
{
	// reload the counter to the required amount of pulses
	TMR1H=0xff;
	TMR1L=cPULSES;

	PIR1=0x00; // the easy way...
	PIE1=0x01; // re-enable interrupt

	TMR1ON=1;
}


//****************************************************************************
//*** EnableTriggerInt
//***
//*** The reverse operation to "EnableCounterInt"
//****************************************************************************
void EnableTriggerInt()
{
	TMR1ON=0;
	PIR1=0x00; // the easy way...
	PIE1=0x08; // re-enable interrupt
}



//****************************************************************************
//*** SendData
//***
//*** ASkr, changes:
//*** - removed function arguments; no stack left, caused random lockups
//****************************************************************************
void SendData()
{
	unsigned int i,k;

	// right now, we are here:
	//
	//      ------+  +--+  +--+  +--+  +--+ 
	// CLK        |  |  |  |  |  |  |  |  | 
	//            +--+  +--+  +--+  +--+  +-
	//               |                      
	//              ISR                     
	//             sens 1                   
	//
	//      ------------+-----+-----+-----+-
	// DAT              |  b0 |  b1 |  b2 | 
	//                  +-----+-----+-----+-
	//                  |<------------------

	k=gMavg;
	
	i=10;
	while(i!=0)
	{
		// wait for falling edge
		while(GP5==1)
		{
			// UNTESTED!
			if(COUT==1)
				return;
		}

		// apply the bit state to the open collector data line
		if((k & 0x01)==0)
			TRISIO&=0xfe;
		else
			TRISIO|=0x01;

		// wait for rising edge
		while(GP5==0)
		{
			// UNTESTED!
			if(COUT==1)
				return;
		}
		
		// next bit
		i--;
		k>>=1;
	}

	// if this is the last sensor in the polling sequence, there will be no
	// further falling edge, which could be used to switch the data line to
	// high impedance state.
	// Therefore, we just wait a fixed amount of time...
	// Potentially unsafe: A future optimizer might throw this out!
	i=3;
	while(i!=0)
		i--;
	
	
	// NOTE:
	// This has bee moved to ISR()!
	// TRISIO|=0x01;
}



//****************************************************************************
//*** ISR
//***
//***
//****************************************************************************
void ISR(void) __interrupt (0)
{
	// Getting here takes about 8us after the interrupt occured (SDCC)

	
	// check for pulse trigger interrupt
	if((PIR1 & 0x08) !=0)
	{
		// wait for rising edge
		while((CMCON0 & 0x40) != 0)
		{;}
		EnableCounterInt();
	}
	else
	{
		// check for pulse count interrupt (timer1 overflow)
		if((PIR1 & 0x01) !=0)
		{
			SendData();
			TRISIO|=0x01;
			EnableTriggerInt();
		}
	}
	
}



//****************************************************************************
//*** main
//***
//***
//****************************************************************************
void main(void)
{
	unsigned int k,l,mavg;
	unsigned char i,j;

	
	// DEBUG ONLY!!!
	unsigned int dbg=0;

	
	Init();

	EnableTriggerInt();
	

	i=0;
	
	for(;;)
	{
		
//	old function call
//	gAvg[i++]=GetAD(10);								

		// -------------------------------------------------
		// --- Get an average of gNumOfSamples AD values
		// --- Do NOT exceed 63 samples!!!
		k=0;
		for(j=0;j<cSAMPLES;j++)							// 25 samples take about 1ms (SDCC)
		{
			GO_DONE=1;
			while(GO_DONE!=0)
			{;}
			k+=ADRESL|(ADRESH<<8);
		}
		gAvg[i++]=k/cSAMPLES;
		if(i>=cAVG)													// avoid errors for running out of retlw area
			i=0;

		// -------------------------------------------------
		// --- create a moving average of gNrOfAvgs values
		// --- Do NOT exceed 63 averages!!!
		mavg=0;															// 5 averages with 25 samples take ~20ms (SDCC)
		for(j=0;j<cAVG;j++)
			mavg+=gAvg[j];
		mavg/=cAVG;
		mavg&=0x3ff;

		
		
		// DEBUG ONLY!
		// mavg=dbg;
		// dbg++;
		// if(dbg>1023)
			// dbg=0;
		
		
		// -------------------------------------------------
		// --- get the corrected values from the array
		//		l=0;
		//		if(mavg != 0) // I can not remember where this check came from... 
//		l=mavg+(mavg>>2);									// get LSB address
//
//		k=Data[l];													// get LSB value
//		j=mavg%4;														// offset for MSB address
//		k+=((Data[l+4-j]<<(j*2+2))&0x300);	// get upper MSBs

		k=Data[mavg];		// get LSB value

	
		EEADR=(mavg/4);	// get MSB value
		RD=1;
		j=mavg%4;
		l=EEDAT;
		k+=((l<<(j*2+2))&0x300);


		// -------------------------------------------------
		// --- Do we need 1:1 values for calibration?
		// --- Writing it this way does not change the
		// --- time behaviour (that much)...
		if(cDOCALIB==0)
			mavg=k;

			
		// -------------------------------------------------
		// --- Additionally, set the PWM (for whatever reason)
		SetPWM(mavg);

		// gMavg will be used during the interrupt routine.
		// because copying this value is not atomic, the interrupts are temporarily disabled
		INTCON&=0x7f;
		gMavg=mavg;
		INTCON|=0x80;

		
	} // end for(;;)

}





