%
% Sharpliner Tool for Octave (and SDCC version of Sharpliner)
% (c) ASkr 10/2008, 12/2009 (DIGITAL EDITION)
% www.askrprojects.com
%
%
%      !!!!!!!!!!!!!!!!!!!!!!!
%      !!! READ THIS FIRST !!! 
%      !!!!!!!!!!!!!!!!!!!!!!!
%
%
% VERSION FOR:
% - Sharpliner DIGITAL >= v0.7, MSBs in EEPROM
%
%
% This tool performs a linearization of a Sharp distance sensor's
%   distance <-> voltage curve.
%
% It will create these files:
%
% data.gif     -> a printout of the calculated curves (as seen on screen)
% data.c       -> a C source code file; copy this to the PIC MPLab (or whatever) project
%                 A description of the coded table can be found in the firmware documentation.
% data.txt     -> a (readable) table of the values in data.c
%
% By default, this script uses a Catmull-Rom Spline interpolation.
%
%
% Put your values, obtained by putting Sharpliner HW in calibration mode,
% in a file named "sharp.dat".
%
% File format:
% <distance in mm> <raw AD value in digits>
% separated by space(s).
% If you need comments in there, use a '%' like this file does.
%
% 100 509
% 150 358
% 200 274
% 250 224
% 300 187
% 350 165
% 400 144
% 450 129
% 500 116
%
%
%
% The digital interface operates with 10 bits, hence the corresponding output values
% will be in a range of 0..1023 (BTW.: The same values are used for the PWM ouput on pin 5;
% Uout=3.3/1024 * digit).
%
% By default (opScale=1), the output data corresponds to the measured distance in mm(!) 1:1.
%
% The opScale factor may be used to fit the output values to suit your needs.
%
% Imagine having a GP2D12 sensor, which is able to measure from 10-80cm.
% An output value of 100 digits represents a distance of 100mm (=10cm).
%
% For an GP2Y0A02 (20-150cm) the standard settings will obviously limit the range to 1023mm.
% Setting opScale to 0.5 results in:
%   1500mm => 750 digits
% Simple, isn't it?
% Note: For performance issues (especially on NXT) your are advised to use 2^n multipliers/divisors.
% (opScale => 2, 4, 8, ... or 0.5, 0.25, 0.125).
% With these settings, the receiving firmware/device can reconstruct human readable values
% with a simple (and fast) shift operation (e.g..: 'AD>>=1', for C).
% But you can set it to whatever value you like...
%%%%%%%%%%%%%%%
opScale=1
%%%%%%%%%%%%%%%



% Linearity Debug Test
% Set to 0 for standard operation (calibration).
% If set to 1, the PWM values will be the reverse of the A/D values, 1:1 (test for linearity).
%   Vout[digits] = 1023 - Vin[digits]; with digits=0..1023
% You won't need this...
%%%%%%%%%%%%%%%
opLinTest=0
%%%%%%%%%%%%%%%



% There might be reasons to limit the distance values. This can be achieved by
% setting opHiLimit and opLoLimit to the required limits.
% (E.g.: You calibrated a sensor that operates on the full range 0..1023.
%  The standard digital interface has no error checking.
%  By setting the minimum and maximum distance values to 1 and 1022, instead of 0 and 1023,
%  your code can check for presence/error of a sensor).
%
% !!! NOTE: The distance will automatically and always be limited to the minimal
% !!! and maximal distance you put into the 'sharp.dat' file!
%
%%%%%%%%%%%%%%%
opHiLimit=1022
opLoLimit=1
%%%%%%%%%%%%%%%



% Title of the document
%%%%%%%%%%%%%%%
opName="SHARPLINER calibration curve"
%%%%%%%%%%%%%%%



% ----------------------------------------------------------------------------
% no (standard) changes below this line
% ----------------------------------------------------------------------------

load sharp.dat

dis=sharp(:,1);
val=sharp(:,2);


% ===================
% REG. POWER FUNCTION
%
% lval=log(val);
% ldis=log(dis);
% pp=polyfit(lval,ldis,1);
% xx=[0:1023]';
% yy=exp(pp(2))*(xx.^pp(1));
% printf("*** RESULTS ***\n");
% printf("***\n");
% printf("*** dis[mm] = %f * ADval[digits] ^ %f\n",exp(pp(2)),pp(1));
% printf("***\n");



% ========================
% REG. CATMULL-ROM SPLINES
%
%pp=spline(val,dis);
pp=catmullrom(val,dis);
xx=[0:1023]';
yy=ppval(pp,xx);


%---------------------------------------------------------------------------------
if(opLinTest==1)
	for i=1:1024
		yy(i)=1024-i;
	endfor
endif



% ============
% LIMIT VALUES
for x=1:1024
	if (x < min(val))
		yy(x)=max(dis);
	endif
	if(x > max(val))
		yy(x)=min(dis);
	endif
	if(yy(x)>max(dis))
		yy(x)=max(dis);
	endif
	if(yy(x)<min(dis))
		yy(x)=min(dis);
	endif
endfor


%---------------------------------------------------------------------------------
yy=yy*opScale;
dis=dis*opScale;


%---------------------------------------------------------------------------------
% just in case someone sets opHi/LoLimit to idiotic values.
% There are even more possibilities. I count on you not to enter crackbrained numbers.
if (opHiLimit>1023)
	opHiLimit=1023;
endif
if (opHiLimit<0)
	opHiLimit=0; % useless, idiot at work  ;)
endif
if (opLoLimit<0)
	opLoLimit=0;
endif
if (opLoLimit>1023)
	opLoLimit=1023; % useless, idiot at work  ;)
endif

for x=1:1024
	if (yy(x)>opHiLimit)
		yy(x)=opHiLimit;
	endif
	if (yy(x)<opLoLimit)
		yy(x)=opLoLimit;
	endif
	
endfor




%---------------------------------------------------------------------------------
subplot(1,1,1);
plot(val,dis,"x","linewidth",3,"color","red",xx,yy,"linewidth",3);
xlabel "measured AD values/digits";
ylabel "distance/mm";
title(opName);
grid "minor" on;
legend "measured digits (scaled)" "calculated curve (scaled)";


%---------------------------------------------------------------------------------
print "-dgif" "-color" "data.gif";




%---------------------------------------------------------------------------------
zz=round(yy);


%---------------------------------------------------------------------------------
fout=fopen("data.c","w+t");

	% --- FLASH ---------------------------------------------------
	fprintf(fout,"\n// created by Sharpliner Octave Tool\n\n\n");
	fprintf(fout,"// LSB-table in flash memory\n");
	fprintf(fout,"__code unsigned char Data[1024]=\n");
	fprintf(fout,"  {");
	
	for i=1:1024

		if(mod(i-1,16)==0)
			fprintf(fout,"\n");
			fprintf(fout,"    ");
		endif

		fprintf(fout,"0x%02x,",bitand(zz(i),0xff));

	endfor
	
	fprintf(fout,"\n  };\n\n");


	% --- EEPROM --------------------------------------------------
	fprintf(fout,"\n\n// MSB-table in eeprom memory\n");
	fprintf(fout,"__code unsigned char __at(0x2100) __EEPROM[256]=\n");
	fprintf(fout,"  {");

	i=1;
	do 

		if(mod(i-1,16*4)==0)
			fprintf(fout,"\n");
			fprintf(fout,"    ");
		endif

		k=0x00;
		for j=0:3
			k=bitshift(k,2,8);
			k+=bitshift(zz(i+j),-8,10);
		endfor
		fprintf(fout,"0x%02x,",k);

		i+=4;
		
	until (i>1023)

	
	fprintf(fout,"\n  };\n\n");
	
	
fclose(fout);


%---------------------------------------------------------------------------------
fout=fopen("data.txt","w+t");
	fprintf(fout,"DIGIT  VALUE\n");
	for i=1:1024
		fprintf(fout," % 5d  % 5d\n",i-1,zz(i));
	end
	fprintf(fout,"\n");
fclose(fout)







%---------------------------------------------------------------------------------

% zz=yy*1000/op1000Scale;
% pp=polyfit(yy,zz,1);
% qq=polyval(pp,xx);

% % *** DISABLED ***
% %zz=(opADref/3.3)*zz;

% for x=1:1024
	% if (zz(x) > 1023)
		% zz(x)=1023;
	% endif
	% if(qq(x) > 1023)
		% qq(x)=1023;
	% endif
% endfor

% subplot(2,1,2)
% plot(yy,zz,"linewidth",3,"color","blue",xx,qq,"linewidth",1,"color","red");
% xlabel "distance/mm";
% ylabel "output PWM/digits";
% grid on;
% legend("sensor distance","extented distance PWM",4);

% print "-dgif" "-color" "data.gif";

% aa=round(zz);



% if(opLinTest==1)
	% for i=1:1024
		% aa(i)=1024-i;
	% endfor
% endif


% fout=fopen("data.c","w+t");

	% fprintf(fout,"\n// created by Sharpliner Octave Tool\n\n");
	% fprintf(fout,"__code unsigned char Data[1280]=\n");
	% fprintf(fout,"  {");
	
	% i=1;
	% do 

		% if(mod(i-1,16)==0)
			% fprintf(fout,"\n");
			% fprintf(fout,"    ");
		% endif

		% k=0x00;
		% for j=0:3
			% k=bitshift(k,2,8);
			% fprintf(fout,"0x%02x,",mod(aa(i+j),256));
			% k+=bitshift(aa(i+j),-8,10);
		% endfor
		% fprintf(fout,"0x%02x,",k);

		% i+=4;
		
	% until (i>1023)
	
	% fprintf(fout,"\n  };\n\n");

% fclose(fout);



% fout=fopen("data.txt","w+t");

	% fprintf(fout,"DIGIT  VALUE\n");

	% for i=1:1024
		% fprintf(fout," % 5d  % 5d\n",i-1,aa(i));
	% end

	% fprintf(fout,"\n");

% fclose(fout)


